@@ -153,6 +153,7 @@ func run() int {
 	flag.BoolVar(&jsMinifier.NoNullishOperator, "js-no-nullish-operator", false, "Don't use the ?? nullish coalescing operator")
 	flag.IntVar(&jsonMinifier.Precision, "json-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
 	flag.BoolVar(&jsonMinifier.KeepNumbers, "json-keep-numbers", false, "Preserve original numbers instead of minifying them")
+	flag.BoolVar(&svgMinifier.KeepComments, "svg-keep-comments", false, "Preserve all comments")
 	flag.IntVar(&svgMinifier.Precision, "svg-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
 	flag.BoolVar(&xmlMinifier.KeepWhitespace, "xml-keep-whitespace", false, "Preserve whitespace characters but still collapse multiple into one")
 	if len(os.Args) == 1 {
@@ -538,6 +539,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 		var info os.FileInfo
 		var err error
 		if !preserveLinks {
+			// follow and dereference symlinks
 			info, err = os.Stat(input)
 		} else {
 			info, err = os.Lstat(input)
@@ -547,7 +549,8 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 		}
 
 		root := filepath.Dir(input) + string(os.PathSeparator) + "."
-		if info.Mode()&os.ModeSymlink != 0 {
+		if preserveLinks && info.Mode()&os.ModeSymlink != 0 {
+			// copy symlink as is
 			if !sync {
 				Warning.Println("--sync not specified, omitting symbolic link", input)
 				continue
@@ -558,7 +561,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 			}
 			tasks = append(tasks, task)
 		} else if info.Mode().IsRegular() {
-			valid := pattern == nil || pattern.MatchString(info.Name())
+			valid := pattern == nil || pattern.MatchString(info.Name()) // don't filter mimetype
 			if valid || sync {
 				task, err := NewTask(root, input, output, !valid)
 				if err != nil {
@@ -609,7 +612,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				}
 
 				if preserveLinks && d.Type()&os.ModeSymlink != 0 {
-					// copy symlinks as is
+					// copy symlink as is
 					if !sync {
 						Warning.Println("--sync not specified, omitting symbolic link", path)
 						return nil
