@@ -56,7 +56,7 @@ func TestHTML(t *testing.T) {
 		{`<span selected="selected"></span>`, `<span selected></span>`},
 		{`<noscript><html><img id="x"></noscript>`, `<noscript><img id=x></noscript>`},
 		{`<body id="main"></body>`, `<body id=main>`},
-		{`<link href="data:text/plain, data">`, `<link href="data:, data">`},
+		{`<link href="data:text/plain, data">`, `<link href=data:,%20data>`},
 		{`<svg width="100" height="100"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow" /></svg>`, `<svg width="100" height="100"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow" /></svg>`},
 		{`<span ></span>`, `<span></span>`},
 		{`</span >`, `</span>`},
@@ -151,17 +151,18 @@ func TestHTML(t *testing.T) {
 		{`<script><!--<`, `<script><!--<`},
 
 		// bugs
-		{`a <span></span> b`, `a <span></span> b`},                                                               // #427
-		{`<i class="fas"></i> Text`, `<i class=fas></i> Text`},                                                   // #390
-		{`<i class="fas"> </i> Text`, `<i class=fas></i>Text`},                                                   // #390
-		{`<span translate="no">`, `<span translate=no>`},                                                         // #352
-		{`<input title=""><input lang="">`, `<input title><input lang>`},                                         // #331
-		{`<amp-analytics type=adobeanalytics_nativeConfig>`, `<amp-analytics type=adobeanalytics_nativeConfig>`}, // #270
-		{`<p>text</p><br>text`, `<p>text</p><br>text`},                                                           // #122
-		{`text <img> text`, `text <img> text`},                                                                   // #89
-		{`text <progress></progress> text`, `text <progress></progress> text`},                                   // #89
-		{`<pre> <x> a  b </x> </pre>`, `<pre> <x> a  b </x> </pre>`},                                             // #82
 		{`<svg id="1"></svg>`, `<svg id="1"></svg>`},                                                             // #67
+		{`<pre> <x> a  b </x> </pre>`, `<pre> <x> a  b </x> </pre>`},                                             // #82
+		{`text <progress></progress> text`, `text <progress></progress> text`},                                   // #89
+		{`text <img> text`, `text <img> text`},                                                                   // #89
+		{`<p>text</p><br>text`, `<p>text</p><br>text`},                                                           // #122
+		{`<amp-analytics type=adobeanalytics_nativeConfig>`, `<amp-analytics type=adobeanalytics_nativeConfig>`}, // #270
+		{`<input title=""><input lang="">`, `<input title><input lang>`},                                         // #331
+		{`<span translate="no">`, `<span translate=no>`},                                                         // #352
+		{`<i class="fas"> </i> Text`, `<i class=fas></i>Text`},                                                   // #390
+		{`<i class="fas"></i> Text`, `<i class=fas></i> Text`},                                                   // #390
+		{`a <span></span> b`, `a <span></span> b`},                                                               // #427
+		{`<canvas><p>test</p></canvas>`, `<canvas><p>test</p></canvas>`},                                         // #440
 	}
 
 	m := minify.New()
@@ -190,8 +191,8 @@ func TestHTMLCSSJS(t *testing.T) {
 		expected string
 	}{
 		// bugs
-		{`<script>var a='<\/script>'</script>`, `<script>var a='<\/script>'</script>`},                                                          // #355
-		{`<script>var a='</scr'+'ipt>'</script>`, `<script>var a='<\/script>'</script>`},                                                        // #355
+		{`<script>var a="<\/script>"</script>`, `<script>var a="<\/script>"</script>`},                                                          // #355
+		{`<script>var a="</scr"+"ipt>"</script>`, `<script>var a="<\/script>"</script>`},                                                        // #355
 		{`<div style="font-family: Arial, &#39;sans-serif&#39;; font-size: 22px;">`, `<div style=font-family:Arial,sans-serif;font-size:22px>`}, // #272
 		{`<style amp-boilerplate>body{-webkit-animation:-amp-start 8s    steps(1,end) 0s 1 normal both;}</style>`, `<style amp-boilerplate>body{-webkit-animation:-amp-start 8s steps(1,end) 0s 1 normal both;}</style>`},
 	}
@@ -278,6 +279,7 @@ func TestHTMLKeepWhitespace(t *testing.T) {
 		{"abc\n</body>\ndef", "abc\ndef"},
 		{"<x>\n<!--y-->\n</x>", "<x>\n</x>"},
 		{"<style>lala{color:red}</style>", "<style>lala{color:red}</style>"},
+		{"<li>one\n</li>\n<li>two", "<li>one\n<li>two"}, // #442
 	}
 
 	m := minify.New()
